function model = smomtl(training, options)
%--------------------------------------------------------------------------
% SMO Sequential Minimal Optimization for SVM based Multi-Task Learning, 
% this algorithm is generalized from John C. Platt's smo algorithm for SVM
% by Feng Cai, Oct.24, 2009
% Prof. Vladimir Cherkassky's machine learning group at the University of 
% Minnesota, Twin Cities.
% 
% Inputs:
%   training
%     training(1), training(2), ... training(t): binary labeled training vectors for 
%     different groups.
%   training(1) [struct] Binary labeled training vectors for group r:
%     .X [num_data x dim] Training vectors.
%     .y [num_data x 1] Labels (1 or -1).
%
%   options [struct] Control parameters:
%     .ker [1x2] Kernel identifier for two spaces (default 0);
%       ker(1) kernel identifier for decision space
%       ker(2) kernel identifier for correction space
%        0 -- linear: u'*v
%        1 -- radial basis function: exp(-|u-v|^2/(2*sigma^2))
%     .arg [1x2] Kernel argument(s) (default 1).
%       arg(1) kernel argument for decision space
%       arg(2) kernel argument for correction space
%     .gamma regularization parameter 
%     .C Regularization constant (default C=inf). 
%
% Output:
%   model [struct] Binary SVM classifier:
%     .Alpha [nsv x 1] Weights (Lagrangians).
%     .groupid [nsv x 1] group index for each alpha
%     .d [tx1] Bias.
%     .svX [nsv x dim] Support vectors.
%     .nsv [1x1] Number of Support Vectors.
%     .options [struct] Copy of used options.

% Input arguments 
% ------------------------------------------------------------------------
if nargin < 2,  options = []; end
if ~isfield(options,'ker')
    options.ker = [0 1];
elseif length(options.ker)~=2
    options.ker = [0 1]; 
end
if ~isfield(options,'arg') 
    options.arg = [1 1]; 
elseif length(options.arg)~=2
    options.arg = [1 1];
end
if ~isfield(options,'gamma')
    options.gamma = 1; 
end
if ~isfield(options,'C')
    options.C = 1; 
end

% run optimizer
%------------------------------------------------------------------------
[model.Alpha,model.groupid,model.d,model.svX,model.svy,model.nsv] =...
       smomtl_mex(training,options.ker,options.arg, options.gamma, options.C );
   
model.options = options;