/* --------------------------------------------------------------------
 smomtl_mex.c: MEX-file for Generalized Sequential Minimal Optimizer.

 Compile:  mex -largeArrayDims smomtl_mex.c

 Synopsis:

  [Alpha,groupid,d,svX,svy,nsv] =
       smomtl_mex(training,ker,arg, gamma, C )

 SMO Sequential Minimal Optimization for SVM based Multi-Task Learning, 
 this algorithm is generalized from John C. Platt's smo algorithm for SVM
 This implementation is modeified from the smo for SVM in STPRTool

 by Feng Cai, Oct. 16, 2009
 Prof. Vladimir Cherkassky's machine learning group at the University of 
 Minnesota, Twin Cities.
 
 Inputs:
   training
     training(1), training(2), ... training(t): binary labeled training vectors for 
	 different groups.
   training(r) [struct] Binary labeled training vectors for group r:
     .X [num_data x dim] Training vectors.
     .y [num_data x 1] Labels (1 or -1).

   
   ker [1x2] Kernel identifier for two spaces (default 0);
	 ker[0] kernel identifier for decision space
     ker[1] kernel identifier for correction space
       0 -- linear: u'*v
       1 -- radial basis function: exp(-|u-v|^2/(2*sigma^2))
   arg [1x2] Kernel argument(s) (default 1).
       arg[0] kernel argument for decision space
       arg[1] kernel argument for correction space
   gamma regularization parameter 
   C Regularization constant (default C=inf). 
   

 Output:
     Alpha [nsv x 1] Weights (Lagrangians).
     groupid [nsv x 1] group index for each alpha
     d [tx1] Bias.
     svX [nsv x dim] Support vectors.
     nsv [1x1] Number of Support Vectors.
*/

#include "mex.h"
#include "matrix.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>


#define ZERO_LIM   1e-9     // patterns with alpha > ZERO_LIM are SV 

#define MAX(A,B)   (((A) > (B)) ? (A) : (B) )
#define MIN(A,B)   (((A) < (B)) ? (A) : (B) )


// --- Global variables ---------------------------------------------- 

unsigned long N = 0;       // number of training patterns 
long DIM;                  // dimension of each instance 
double C;				   // trade-off constants 
double GAMMA;              // Regularization parameter
double tolerance=0.001;    // tolerance in KKT fulfilment  
double eps=0.001;          // minimal Lagrangeian change 
double **DATA;              // pointer at training instances 
double *TARGET;            // pointer at labels 
int *GROUPID;              // group id for each instance 
double *ALPHA;             // Lagrange multipliers 
double *D;                 // Bias (threshold) for each group 
double *ERROR_CACHE;       // Store the error cache
int KER[2];                // KER[0] kernel for decision space
						   //      0: linear kernel
						   //   	 1: rbf kernel
						   // KER[1] kernel for correction space
						   //     0: linear kernel
						   //    1: rbf kernel 
double ARG[2];             // ARG[0] parameter of kernel for decision space
						   //   ARG[1] parameter of kernel for correction space 


// ==============================================================
// Implementation of Generalized Sequential Minimal Optimizer (GSMO)
// ============================================================ 

// --------------------------------------------------------------
// Kernel funtion for mtl
// -------------------------------------------------------------- 
double mtlkernel(long i1, long i2, int ker, double arg)
{
	double c = 0.;
	long i;
	if( ker==0 )
	{
		// linear kernel
		for( i=0; i<DIM; i++ )
		{
			c = c + DATA[i1][i]*DATA[i2][i];
		}
	}
	else
	{
		// rbf kernel
		for( i=0; i<DIM; i++ )
		{
			c = c + (DATA[i1][i]-DATA[i2][i])*(DATA[i1][i]-DATA[i2][i]);
		}
		c = exp(-0.5*c/(arg*arg));
	}
	return c;
}

// --------------------------------------------------------------
// Computes value of the learned function for k-th instance.
// ------------------------------------------------------------ 
double learned_func( long k )
{
	double s = 0.;
	long i;
	for( i=0; i<N; i++ )
	{
		if( ALPHA[i]>0 )
		{
			s = s + ALPHA[i]*TARGET[i]*mtlkernel( i, k, KER[0], ARG[0] );

			if( GROUPID[i]==GROUPID[k] )
			{
				s = s + 1/GAMMA*ALPHA[i]*TARGET[i]*mtlkernel( i, k, KER[1], ARG[1] );
			}
		}
	}
	s = s + D[GROUPID[k]-1];
	return s;
}


// --------------------------------------------------------------
// Optimizes objective function for i1-th and i2-th instance.
// ------------------------------------------------------------- 
long takeStep( long i1, long i2 )
{
	double y1, y2, s;
    long i;
    double alpha1, alpha2; 
    double a1, a2;
    double E1, E2, L, H, L1, H1;
	double k11, k22, k12, kr11, kr12, kr22, eta, Lobj, Hobj;
    double f1, f2, k1i, kr1i, k2i, kr2i;
    double t1, t2;
    double d1, d2, dnew;
    double delta_d;
    
	if( i1==i2 )
	{
		return 0;
	}

	alpha1 = ALPHA[i1];
	y1 = TARGET[i1];
	if( alpha1>0 && alpha1<C )
	{
		E1 = ERROR_CACHE[i1];
	}
	else
	{
		E1 = learned_func(i1)-y1;
	}

	alpha2 = ALPHA[i2];
	y2 = TARGET[i2];
	if( alpha2>0 && alpha2<C )
	{
		E2 = ERROR_CACHE[i2];
	}
	else
	{
		E2 = learned_func(i2)-y2;
	}

	s = y1 * y2;

	if( s<0 )
	{
		L = MAX(0, alpha2-alpha1);
		H = MIN(C, C+alpha2-alpha1);
	}
	else
	{
		L = MAX(0, alpha2+alpha1-C);
		H = MIN(C, alpha2+alpha1);
	}

	if( L==H )
	{
		return 0;
	}

	k11 = mtlkernel(i1, i1, KER[0], ARG[0]);
	kr11 = mtlkernel(i1, i1, KER[1], ARG[1]);
	k12 = mtlkernel(i1, i2, KER[0], ARG[0]);
	kr12 = mtlkernel(i1, i2, KER[1], ARG[1]);
	k22 = mtlkernel(i2, i2, KER[0], ARG[0]);
	kr22 = mtlkernel(i2, i2, KER[1], ARG[1]);

	eta = k11 + 1/GAMMA*kr11 -2*k12 - 2/GAMMA*kr12 + k22 + 1/GAMMA*kr22;

	if( eta>0 )
	{
		a2 = alpha2 + y2*(E1-E2)/eta;
		if( a2<L )
		{
			a2 = L;
		}
		else if( a2>H )
		{
			a2 = H;
		}
	}
	else
	{
		f1 = y1*(E1-D[GROUPID[i1]-1]) - alpha1*k11 - s*alpha2*k12 - 1/GAMMA*alpha1*kr11 - 1/GAMMA*s*alpha2*kr12;
		f2 = y2*(E2-D[GROUPID[i2]-1]) - alpha2*k22 - s*alpha1*k12 - 1/GAMMA*alpha2*kr22 - 1/GAMMA*s*alpha1*kr12;
		L1 = alpha1 + s*(alpha2-L);
		H1 = alpha1 + s*(alpha2-H);
		Lobj = L1*f1 + L*f2 + 0.5*L1*L1*k11 + 0.5*L*L*k22 + s*L1*L*k12 + 1/GAMMA*(0.5*L1*L1*kr11 + 0.5*L*L*kr22 + s*L1*L*kr12);
		Hobj = H1*f1 + H*f2 + 0.5*H1*H1*k11 + 0.5*H*H*k22 + s*H1*H*k12 + 1/GAMMA*(0.5*H1*H1*kr11 + 0.5*H*H*kr22 + s*H1*H*kr12);
		if( Lobj>Hobj+eps )
		{
			a2 = H;
		}
		else if( Lobj<Hobj-eps )
		{
			a2 = L;
		}
		else
		{
			a2 = alpha2;
		}
	}
	
	if( fabs(a2-alpha2) < eps*(a2+alpha2+eps )) 
	{
		return( 0 );
	}

	a1 = alpha1 + s*(alpha2 - a2);
	if( a1<0 )
	{
		a2 = a2 + s*a1;
		a1 = 0;
	}
	else if( a1>C )
	{
		a2 = a2 + s*(a1-C);
		a1 = C;
	}

	if( a1>0 && a1<C )
	{
		dnew = D[GROUPID[i1]-1] - E1 - y1*(a1-alpha1)*k11 -y2*(a2-alpha2)*k12 - 1/GAMMA*(y1*(a1-alpha1)*kr11 + y2*(a2-alpha2)*kr12);
	}
	else
	{
		if( a2>0 && a2<C )
		{
			dnew = D[GROUPID[i2]-1] - E2 - y1*(a1-alpha1)*k12 - y2*(a2-alpha2)*k22 - 1/GAMMA*(y1*(a1-alpha1)*kr12 + y2*(a2-alpha2)*kr22);
		}
		else
		{
			d1 = D[GROUPID[i1]-1] - E1 - y1*(a1-alpha1)*k11 -y2*(a2-alpha2)*k12 - 1/GAMMA*(y1*(a1-alpha1)*kr11 + y2*(a2-alpha2)*kr12);
			d2 = D[GROUPID[i2]-1] - E2 - y1*(a1-alpha1)*k12 - y2*(a2-alpha2)*k22 - 1/GAMMA*(y1*(a1-alpha1)*kr12 + y2*(a2-alpha2)*kr22);
			dnew = (d1 + d2)/2;
		}
	}

	delta_d = dnew - D[GROUPID[i1]-1];
	D[GROUPID[i1]-1] = dnew;
	t1 = y1 * (a1-alpha1);
	t2 = y2 * (a2-alpha2);

	for( i=0; i<N; i++ )
	{
		if( ALPHA[i]>0 && ALPHA[i]<C )
		{
			k1i = mtlkernel(i1, i, KER[0], ARG[0]);
			kr1i = mtlkernel(i1, i, KER[1], ARG[1]);
			k2i = mtlkernel(i2, i, KER[0], ARG[0]);
			kr2i = mtlkernel(i2, i, KER[1], ARG[1]);
			if( GROUPID[i]==GROUPID[i1] )
			{
				ERROR_CACHE[i] = ERROR_CACHE[i] + t1*k1i + t2*k2i + 1/GAMMA*(t1*kr1i + t2*kr2i) + delta_d;
			}
			else
			{
				ERROR_CACHE[i] = ERROR_CACHE[i] + t1*k1i + t2*k2i;
			}
		}
	}

	ERROR_CACHE[i1] = 0;
	ERROR_CACHE[i2] = 0;
	ALPHA[i1] = a1;
	ALPHA[i2] = a2;

	return 1;
}



// --------------------------------------------------------------
// Finds the second Lagrange multiplayer to be optimize.
// ------------------------------------------------------------- 
long examineExample( long i1 )
{
	double y1, alpha1, E1, r1;
	double tmax;
	double E2, temp;
	long i, i2;
	long imin, imax;
	
	alpha1 = ALPHA[i1];
	y1 = TARGET[i1];
	if( alpha1>0 && alpha1<C )
	{
		E1 = ERROR_CACHE[i1];
	}
	else
	{
		E1 = learned_func(i1) - y1;
	}

	r1 = y1 * E1;
	if( (r1<-1*tolerance && alpha1<C) || (r1>tolerance && alpha1>0) )
	{
		imin = i1;
		while( imin>=0 && GROUPID[i1]==GROUPID[imin] )
		{
			imin = imin - 1;
		}
		imin = imin + 1;

		imax = i1;
		while( imax<N && GROUPID[i1]==GROUPID[imax] )
		{
			imax = imax + 1;
		}
		imax = imax - 1;

		i2 = -1;
		tmax = 0;

		for( i=imin; i<=imax; i++ )
		{
			if( ALPHA[i]>0 && ALPHA[i]<C )
			{
				E2 = ERROR_CACHE[i];
				temp = fabs(E1-E2);
				if( temp>tmax )
				{
					tmax = temp;
					i2 = i;
				}
			}
		}

		if( i2>-1 )
		{
			if( takeStep(i1,i2) )
			{
				return 1;
			}
		}

		for( i2 = imin; i2<=imax; i2++ )
		{
			if( ALPHA[i2]>0 && ALPHA[i2]<C )
			{
				if( takeStep(i1, i2) )
				{
					return 1;
				}
			}
		}

		for( i2=imin; i2<=imax; i2++ )
		{
			if( takeStep(i1, i2) )
			{
				return 1;
			}
		}
	}

	return 0;
}



// --------------------------------------------------------------
// Main SMO optimization cycle.
// ------------------------------------------------------------ 
void runSMOMTL( void )
{
	long numChanged = 0;
	long examineAll = 1;
	long i;

	while( numChanged>0 || examineAll )
	{
		
		numChanged = 0;

		if( examineAll )
		{
			for( i=0; i<N; i++ )
			{
				numChanged = numChanged + examineExample( i );
			}
		}
		else
		{
			for( i=0; i<N; i++ )
			{
				if( ALPHA[i]!=0 && ALPHA[i]!=C )
				{
					numChanged = numChanged + examineExample( i );
				}
			}
		}

		if( examineAll==1 )
		{
			examineAll = 0;
		}
		else if( numChanged==0 )
		{
			examineAll = 1;
		}
	}
}



// ==============================================================
// Main MEX function - interface to Matlab.
// ============================================================ 

void mexFunction( int nlhs, mxArray *plhs[],
		  int nrhs, const mxArray*prhs[] )
{
	
	long i,j,k,t;
	long i1, i2;
	long r;
	double *nsv;
	double C1, C2;
	mwSize NStructElems;
	mwSize     ndim;
	mxArray    *tmp, *tmp2;
	double *arraytmp, *arraytmp2, *arraytmp3, *arraytmp4;
	N = 0;
	

	//---- get input arguments ------------------------
	if( nrhs!=5 )
	{
		mexErrMsgTxt("5 inputs required.");
	}
	else if( nlhs!=6 )
	{
		mexErrMsgTxt("6 outputs required.");
	}

	if( !mxIsStruct(prhs[0]) )
	{
		mexErrMsgTxt("Input must be a structure.");
	}

    NStructElems = mxGetNumberOfElements(prhs[0]);

	for( i=0; i<NStructElems; i++ )
	{
		tmp = mxGetField(prhs[0], i, "X");
		DIM = mxGetN(tmp);
		j = mxGetM(tmp);
		N = N + j;
	}
	
	// allocate memory for DATA
	if( (DATA = mxCalloc(N, sizeof(*DATA) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory.");
	}
	for( i=0; i<N; i++ )
	{
		if( (DATA[i] = mxCalloc(DIM, sizeof(double) )) == NULL) 
		{
			mexErrMsgTxt("Not enough memory.");
		}
	}

	// allocate memory for TARGET (labels) (1,-1) 
	if( (TARGET = mxCalloc(N, sizeof(double) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory.");
	}

	// allocate memory for GROUPID
	if( (GROUPID = mxCalloc(N, sizeof(int) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory.");
	}

	// allocate memory for ALPHA
	if( (ALPHA = mxCalloc(N, sizeof(double) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory for error cache.");
	}

	// allocate memory for ERROR_CACHE 
	if( (ERROR_CACHE = mxCalloc(N, sizeof(double) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory for error cache.");
	}

	k = 0;
	for( i=0; i<NStructElems; i++ )
	{
		tmp = mxGetField(prhs[0], i, "X");
		DIM = mxGetN(tmp);
		j = mxGetM(tmp);
		arraytmp = mxGetPr(tmp);
		tmp2 = mxGetField(prhs[0], i, "y");
		arraytmp2 = mxGetPr(tmp2);
		
		for( i1=0; i1<j; i1++ )
		{
			for( i2=0; i2<DIM; i2++ )
			{
				DATA[i1+k][i2] = arraytmp[i2*j+i1];
			}

			GROUPID[i1+k] = i+1;
			TARGET[i1+k] = arraytmp2[i1];
			ALPHA[i1+k] = 0;
			ERROR_CACHE[i1+k] = 0;
		}
		k = k + j;
	}

	r = i;
	// allocate memory for D
	if( (D = mxCalloc(r, sizeof(double) )) == NULL) 
	{
		mexErrMsgTxt("Not enough memory for error cache.");
	}
	for( i=0; i<r; i++ )
	{
		D[i] = 0;
	}

	// get value for KER
	arraytmp = mxGetPr( prhs[1] );
	KER[0] = arraytmp[0];
	KER[1] = arraytmp[1];

	// get value for ARG
	arraytmp = mxGetPr( prhs[2] );
	ARG[0] = arraytmp[0];
	ARG[1] = arraytmp[1];

	// get value for GAMMA
	arraytmp = mxGetPr( prhs[3] );
	GAMMA = *arraytmp;

	// get value for C
	arraytmp = mxGetPr( prhs[4] );
	C = *arraytmp;

	//---------- run SMOMTL ---------------------------------------------
	runSMOMTL( );


	//---------- outputs ------------------------------------------------

	// assign value for nsv: number of support vectors
	plhs[5] = mxCreateDoubleMatrix(1,1,mxREAL);
	arraytmp = mxGetPr( plhs[5] );
	*arraytmp = 0;
	for( i=0; i<N; i++ )
	{
		if( ALPHA[i] > ZERO_LIM )
		{
			*arraytmp = *arraytmp + 1;
		}
		else
		{
			ALPHA[i] = 0;
		}
	}
	k = *arraytmp;

	// assign value for Alpha, groupid, svX
	plhs[0] = mxCreateDoubleMatrix(k,1,mxREAL);
	arraytmp = mxGetPr( plhs[0] );
	plhs[1] = mxCreateDoubleMatrix(k,1,mxREAL);
	arraytmp2 = mxGetPr( plhs[1] );
	plhs[3] = mxCreateDoubleMatrix(k,DIM,mxREAL);
	arraytmp3 = mxGetPr( plhs[3] );
	plhs[4] = mxCreateDoubleMatrix(k,1,mxREAL);
	arraytmp4 = mxGetPr( plhs[4] );
	
	j = 0;
	for( i=0; i<N; i++ )
	{
		if( ALPHA[i] > ZERO_LIM )
		{
			// Alpha
			arraytmp[j] = ALPHA[i]; 	

			// groupid
			arraytmp2[j] = GROUPID[i];

			// svX
			for( t=0; t<DIM; t++ )
			{
				arraytmp3[t*k+j] = DATA[i][t];
			}

			// svy
			arraytmp4[j] = TARGET[i];
			j = j + 1;
		}
	}

	// assign value for d
	plhs[2] = mxCreateDoubleMatrix(r,1,mxREAL);
	arraytmp = mxGetPr( plhs[2] );
	for( i=0; i<r; i++ )
	{
		arraytmp[i] = D[i];
	}

	for( i=0; i<N; i++)
	{
		mxFree( DATA[i] );
	}
	mxFree( (void *)DATA );
	mxFree( TARGET );
	mxFree( ERROR_CACHE );
	mxFree( ALPHA );
	mxFree( GROUPID );
	mxFree( D );
}